#=======================================================================
#	$Id: ShellCmd.pm,v 1.7 2013/09/30 15:09:00 pythontech Exp $
#	Command line utility for manipulating triplestore
#	Copyright (C) 2005-2007  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#-----------------------------------------------------------------------
#	E.g. when used in a command '3util':
#
#	3util $db info
#	3util $db query x y z
#=======================================================================
package PythonTech::Triples::ShellCmd;
use strict;

#-----------------------------------------------------------------------
#	Create e new instance
#	  triples	Triple store
#	  name		Application name for error reports
#	  howused	Preamble for usage messages (optional)
#	  lock		True if exclusive lock wanted
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {@_};
    $self->{'triples'} ||= $self->{'db'};
    die "$class: triples (or db) not configured\n" unless $self->{'triples'};
    bless $self, $class;
    return $self;
}

my $cmdmap = {
    'info' => [\&_info], 
    'configure' => [\&_configure, '<attribute>','<value>'], 
    'ntriples' => [\&_ntriples], 
    'load-ntriples' => [\&_load_ntriples, '[<filename>]'], 
    'list' => [\&_list, '<type>'],
    'show' => [\&_show, '<subj>'],
    'query' => [\&_query, '<subj>','<pred>','<obj>'],
    'query3' => [\&_query3, '<subj>','<pred>','<obj>'],
    'add' => [\&_add, '<subj>','<pred>','<obj>'],
    'set' => [\&_set, '<subj>','<pred>','<obj>'],
    'remove' => [\&_remove, '<subj>','<pred>','<obj>'],
    'create' => [\&_create, '<subj>','<type>','[<pred> <obj>]...'],
    'destroy' => [\&_destroy, '<subj>'],
    'rename' => [\&_rename, '<olduri>','<newuri>'],
    'dump' => [\&_dump],
};

sub usage {
    my $self = shift;
    my $myname = $self->{'howused'} || $self->{'name'};
    print STDERR ("Usage: ",
		  join("\n       ",
		       map {
			   my($f,@a) = @{$cmdmap->{$_}};
			   join(" ",$myname,$_,@a);
		       }
		       sort
		       keys %$cmdmap),
		  "\n");
    exit 200;
}

#-----------------------------------------------------------------------
#	Main program
#-----------------------------------------------------------------------
sub main {
    my($self,$myname,@args) = @_;
    $self->{'name'} ||= $myname;

    # Get method
    @args >= 1 || $self->usage;
    my $cmd = shift @args;
    my $desc = $cmdmap->{$cmd} || $self->usage;
    my($func,@adesc) = @$desc;
    # Check argument count
    my @opt = grep {/^\[.*\]$/} @adesc;
    my @any = grep {/\.\.\.$/} @adesc;
#    print "args=",scalar @args," adesc=",scalar @adesc," opt=",scalar @opt," any=",scalar @any,"\n";
    if (@any) {
	@args >= @adesc - @opt - @any || $self->usage;
    } else {
	@args >= @adesc - @opt || $self->usage;
	@args <= @adesc || $self->usage;
    }
    # Call method
    $func->($self,$self->{'triples'},@args);
}

sub _info {
    my($self,$triples) = @_;
    print "class: ",ref($triples),"\n";
    if ($triples->can('metadata')) {
	foreach my $meta (sort $triples->metadata) {
	    print "$meta: ",join(" ",$triples->metadata($meta)),"\n";
	}
    }
}

sub _configure {
    my($self,$triples, $attr,$val) = @_;
    $self->_writable;
    $triples->configure($attr => $val);
    $triples->commit;
}

sub _ntriples {
    my($self,$triples) = @_;
    $triples->write_ntriples();
}

sub _load_ntriples {
    my($self,$triples,$filename) = @_;
    $self->_writable;
    if (! defined $filename) {
	$triples->read_ntriples(\*STDIN);
    } else {
	open(NT,'<',$filename) || die "Cannot open $filename: $!\n";
	$triples->read_ntriples(\*NT);
	close(NT);
    }
    $triples->commit;
}

sub _list {
    my($self,$triples, $type) = @_;
    foreach my $subj ($triples->po2s('rdf:type', $type)) {
	print "$subj\n";
    }
}

sub _dump {
    my($self,$triples) = @_;
    foreach my $stmt ($triples->query(undef,undef,undef)) {
	my($subj,$pred,$obj) = @$stmt;
	if ($obj =~ /^=/) {
	    print "$subj $pred ",&shellquote($obj),"\n";
	} else {
	    print "$subj $pred $obj\n";
	}
    }
}

sub _show {
    my($self,$triples, $subj) = @_;
    foreach my $pred ($triples->s2p($subj)) {
	foreach my $obj ($triples->sp2o($subj,$pred)) {
	    print "$pred\t$obj\n";
	}
    }
}

sub _query {
    my($self,$triples, $ts,$tp,$to) = @_;
    my($subj,$pred,$obj) = map {$_ eq '?' ? undef : $_} $ts,$tp,$to;
    $triples->iterate($subj,$pred,$obj, sub {
	my($s,$p,$o) = @_;
	print join(" ",
		   defined $subj ? () : ($s),
		   defined $pred ? () : ($p),
		   defined $obj ? () : ($o),
		   ),"\n";
	return;
    });
}

sub _query3 {
    my($self,$triples, $ts,$tp,$to) = @_;
    my($subj,$pred,$obj) = map {$_ eq '?' ? undef : $_} $ts,$tp,$to;
    $triples->iterate($subj,$pred,$obj, sub {
	print join(" ",@_),"\n";
	return;
    });
}

sub _add {
    my($self,$triples, $subj,$pred,$obj) = @_;
    $self->_writable;
    $triples->add($subj,$pred,$obj);
    $triples->commit;
}

sub _set {
    my($self,$triples, $subj,$pred,$obj) = @_;
    $self->_writable;
    foreach my $o ($triples->sp2o($subj,$pred)) {
	$triples->remove($subj,$pred,$o);
    }
    $triples->add($subj,$pred,$obj);
    $triples->commit;
}

sub _remove {
    my($self,$triples, $subj,$pred,$obj) = @_;
    $self->_writable;
    my $there = $triples->test($subj,$pred,$obj);
    if ($there) {
	$triples->remove($subj,$pred,$obj);
	$triples->commit;
    } else {
	print STDERR "Was not set: $subj $pred $obj\n";
    }
}

sub _create {
    my($self,$triples, $subj,$class,@po) = @_;
    $self->_writable;
    @po%2==0 || $self->usage;
    $triples->add($subj,'rdf:type',$class);
    while (my($pred,$obj) = splice(@po,0,2)) {
	$triples->add($subj,$pred,$obj);
    }
    $triples->commit;
}

#    } elsif ($cmd eq 'destroy') {	# destroy uri
#	die "$0: 'destroy' method unimplemented\n";
#	(@args == 1) || $self->usage;
#	my($subj) = @args;
#	foreach my $po ($triples->query($subj,undef,undef)) {
#	    $triples->remove($subj,@$po);
#	}
#	foreach my $so ($triples->query(undef,$subj,undef)) {
#	    my($s,$o) = @$so;
#	    $triples->remove($s,$subj,$o);
#	}
#	foreach my $sp ($triples->query(undef,undef,$subj)) {
#	    $triples->remove(@$sp,$subj);
#	}
#	$triples->commit;

sub _rename {
    my($self,$triples, $from,$to) = @_;
    $self->_writable;
    foreach my $po ($triples->query($from,undef,undef)) {
	$triples->remove($from,@$po);
	$triples->add($to,@$po);
    }
    foreach my $so ($triples->query(undef,$from,undef)) {
	my($s,$o) = @$so;
	$triples->remove($s,$from,$o);
	$triples->add($s,$to,$o);
    }
    foreach my $sp ($triples->query(undef,undef,$from)) {
	$triples->remove(@$sp,$from);
	$triples->add(@$sp,$to);
    }
    $triples->commit;
}

#-----------------------------------------------------------------------
#	Re-open the triplestore with write access
#-----------------------------------------------------------------------
sub _writable {
    my $self = shift;
    my $mode = 'w';
    $mode .= 'l' if $self->{'lock'};
    eval {$self->{'triples'}->reconnect($mode), 1} or
	die "$self->{'name'}: Cannot reopen for write: $@\n";
}

sub shellquote {
    my($text) = @_;
    $text =~ s/([\$\"\`\\])/\\$1/g;
    $text =~ s/\r/\${cr}/g;
    $text =~ s/\n/\${nl}/g;
    return '"' . $text . '"';
}

1;
